/*
 * linux.h
 * Copyright (C) 2003-2016, Gerdes Aktiengesellschaft. All rights reserved.
 *
 * Contact: Gerdes AG, Bergerwiesenstr. 9, 53340 Meckenheim, Germany, email: info@gerdes-ag.de
 */

#ifndef _LINUX_LIB_H_
#define _LINUX_LIB_H_

#if defined(_DRV_LINUX_)

#if defined(__cplusplus)
extern "C" {
#endif

// Attribut fr eigene Funktionen
#define LINUXAPI            __attribute__((regparm(3)))

// Datentypen
#ifdef NULL
#undef NULL
#endif
#define NULL                __null

#ifndef LOBYTE
#define LOBYTE(l)           ((BYTE)(l))
#define HIBYTE(l)           ((BYTE)(((WORD)(l) >> 8) & 255))
#define LOWORD(l)           ((WORD)(l))
#define HIWORD(l)           ((WORD)(((DWORD)(l) >> 16) & 0xFFFF))
#define MAKELONG(a, b)      ((LONG)(((WORD)(a)) | ((DWORD)((WORD)(b))) << 16))
#endif

typedef unsigned char       BYTE;
typedef unsigned short      WORD;
typedef unsigned int        DWORD;
typedef unsigned int        UINT;
typedef unsigned int 	    LONG;
typedef unsigned int        ULONG;
typedef int                 BOOL, BOOLEAN;
typedef int                 STATUS;

typedef void*               LPVOID;
typedef BYTE*               LPBYTE;
typedef int*                LPINT;
typedef WORD*               LPWORD;
typedef LONG*               LPLONG;
typedef DWORD*              LPDWORD;
typedef char*               LPSTR;
typedef const char*         LPCSTR;
typedef const BYTE*         LPCBYTE;
typedef const void*         LPCVOID;

#define TRUE                1
#define FALSE               0

// Hardware-Verwaltung
#define STATUS_SUCCESS					   0
#define STATUS_FAILURE					  -1
#define STATUS_PENDING					   1
#define STATUS_CANCELLED				   2
#define STATUS_NOT_SUPPORTED			  -25
#define STATUS_INVALID_PARAMETER		  -22
#define ERROR_INVALID_PARAMETER		  -22
#define ERROR_INSUFFICIENT_BUFFER     -12
#define ERROR_UNEXPECTED_CALL         -5
#define ERROR_REG_DEVFILE             -6
#define ERROR_REG_DEVDRV              -7
#define STATUS_INSUFFICIENT_RESOURCES -12
#define STATUS_TOO_MANY_SESSIONS      -5
#define STATUS_CAPI_ERROR             -5
#define STATUS_NO_SUCH_DEVICE         -6
#define STATUS_HALTED                 -7
#define STATUS_DATA_OVERRUN           -8

#define ERROR_REG_PCI                 0x0001
#define ERROR_REG_USB                 0x0002
#define ERROR_REG_PNP                 0x0004

#define DEV_TYPE_PCI                  0x01000000
#define DEV_TYPE_USB                  0x02000000
#define DEV_TYPE_ISA                  0x04000000
#define DEV_TYPE_PNP                  0x08000000

typedef struct _SUPHWDEV {
	WORD		id;
   DWORD		type;
	UINT		vendor;
	UINT		subvendor;
	UINT		device;
	UINT		subdevice;
	ULONG		drv_data;
	WORD		friendlyName;
	LPCSTR	name;
	} SUPHWDEV, *PSUPHWDEV, *LPSUPHWDEV;

typedef struct _PARMSTRUCT {
	WORD        irq, comPort, window, socket;
	WORD        port1, port2;
	DWORD       mem1,  mem2, mem3;
	LPBYTE      ptr1,  ptr2, ptr3;
	} PARMSTRUCT, *PPARMSTRUCT, *LPPARMSTRUCT;

typedef struct _HARDW_DEV {
	DWORD				devID;
	DWORD          devAddr[2];
	PARMSTRUCT	   devParms;
	LPSUPHWDEV		pModel;
	LPVOID			pOsDev;
	} HARDW_DEV, *PHARDW_DEV, *LPHARDW_DEV;

typedef union _DWORD_PTR {
	DWORD  dw;
	LPVOID ptr;
	} DWORD_PTR;

extern LPVOID  LINUXAPI LPmRegisterDev(LPVOID irpHdl, DWORD devType, DWORD id);
extern void    LINUXAPI LPmUnregisterDev(LPVOID pmHdl);
extern void    LINUXAPI LPnpCleanup(LPVOID pDev);
extern void    LINUXAPI LPciConfigWrite(LPVOID pDev, LPVOID buf, DWORD ofs, DWORD len);
extern void    LINUXAPI LPciConfigRead(LPVOID pDev, LPVOID buf, DWORD ofs, DWORD len);
extern STATUS  LINUXAPI LRegSupHwDev(const PSUPHWDEV dev_tab, LPCSTR drv_name);
extern STATUS  LINUXAPI LUnregSupHwDev(void);

///////////////////////////////////////////////////////////////////////////////
// Listenverwaltung
typedef struct _list_entry {
	struct _list_entry*   next;
	struct _list_entry*   prev;
	} LIST_ENTRY, *PLIST_ENTRY;

#define CONTAINING_RECORD(type, member, ptr)   ((type*)((char*)(ptr) - ((size_t)&((type*)0)->member)))

extern void        LINUXAPI LListInit(PLIST_ENTRY list);
extern void        LINUXAPI LListAddTail(PLIST_ENTRY list, PLIST_ENTRY ent);
extern void        LINUXAPI LListRemoveEntry(PLIST_ENTRY ent);
extern PLIST_ENTRY LINUXAPI LListRemoveHead(PLIST_ENTRY list);
extern int         LINUXAPI LListIsEmpty(PLIST_ENTRY list);

///////////////////////////////////////////////////////////////////////////////
// USB
struct urb;
union _urb;

#define USBD_TRANSFER_DIRECTION_IN          0x01
#define USBD_SHORT_TRANSFER_OK              0x02

#define USB_ENDPOINT_DIRECTION_MASK         0x80
#define USB_ENDPOINT_DIRECTION_IN(addr)     ((addr) & USB_ENDPOINT_DIRECTION_MASK)
#define USB_ENDPOINT_DIRECTION_OUT(addr)    (!((addr) & USB_ENDPOINT_DIRECTION_MASK))

typedef enum _USBD_PIPE_TYPE {
	UsbdPipeTypeControl,
	UsbdPipeTypeIsochronous,
	UsbdPipeTypeBulk,
	UsbdPipeTypeInterrupt,
	} USBD_PIPE_TYPE;

typedef struct USBD_PIPE_INFORMATION {
	unsigned short        MaximumPacketSize;
	unsigned char         EndpointAddress;
	unsigned char         Interval;
	USBD_PIPE_TYPE        PipeType;
	unsigned int          PipeHandle;
	} *PUSBD_PIPE_INFORMATION;

typedef enum _URB_FUNCTION {
	URB_FUNCTION_GET_CURRENT_FRAME_NUMBER,
	URB_FUNCTION_ABORT_PIPE,
	URB_FUNCTION_RESET_PIPE,
	URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER,
	URB_FUNCTION_ISOCH_TRANSFER,
	URB_FUNCTION_CONTROL_TRANSFER,
	} URB_FUNCTION;

typedef struct _URB_HEADER {
	LIST_ENTRY            Entry;
	enum _URB_FUNCTION    Function;
	unsigned int          Length;
	unsigned int          Status;
	LONG                  Ref;
	void                  (*Complete)(union _urb*) LINUXAPI;
	void*                 Context;
	struct urb*           Urb;
	} URB_HEADER;

typedef struct _URB_PIPE_REQUEST {
	struct _URB_HEADER    Header;
	unsigned int          PipeHandle;
	// ...
	} URB_PIPE_REQUEST;

typedef struct _URB_GET_CURRENT_FRAME_NUMBER {
	struct _URB_HEADER    Header;
	unsigned int          PipeHandle;
	unsigned int          FrameNumber;
	// ...
	} URB_GET_CURRENT_FRAME_NUMBER;

typedef struct _URB_ISO_PKT {
	unsigned short        Offset;
	unsigned short        Length;
	unsigned int          Status;
	} URB_ISO_PKT;

typedef struct _URB_ISOCHRONOUS_TRANSFER {
	struct _URB_HEADER    Header;
	unsigned int          PipeHandle;
	unsigned int          Interval;
	void*                 TransferBuffer;
	unsigned int          TransferBufferLength;
	unsigned int          TransferFlags;
	unsigned int          StartFrame;
	unsigned int          NumberOfPackets;
	unsigned int          ErrorCount;
	struct _URB_ISO_PKT   IsoPacket[0];
	// ...
	} URB_ISOCHRONOUS_TRANSFER;

typedef struct _URB_BULK_OR_INTERRUPT_TRANSFER {
	struct _URB_HEADER    Header;
	unsigned int          PipeHandle;
	unsigned int          Interval;
	void*                 TransferBuffer;
	unsigned int          TransferBufferLength;
	unsigned int          TransferFlags;
	unsigned int          ActualLength;
	// ...
	} URB_BULK_OR_INTERRUPT_TRANSFER;

typedef struct _URB_CONTROL_TRANSFER {
	struct _URB_HEADER    Header;
	void*                 TransferBuffer;
	unsigned int          TransferBufferLength;
	unsigned int          TransferFlags;
	unsigned int          ActualLength;
	unsigned char         RequestType;
	unsigned char         Request;
	unsigned short        Value;
	unsigned short        Index;
	} URB_CONTROL_TRANSFER;

typedef union _urb {
	struct _URB_HEADER                          UrbHeader;
	struct _URB_PIPE_REQUEST                    UrbPipeRequest;
	struct _URB_GET_CURRENT_FRAME_NUMBER        UrbGetCurrentFrameNumber;
	struct _URB_ISOCHRONOUS_TRANSFER            UrbIsochronousTransfer;
	struct _URB_BULK_OR_INTERRUPT_TRANSFER      UrbBulkOrInterruptTransfer;
	struct _URB_CONTROL_TRANSFER                UrbControlTransfer;
	} URB, *PURB;

extern LPVOID  LINUXAPI LUsbAllocBuffer(int size);
extern void    LINUXAPI LUsbFreeBuffer(LPVOID ptr);

extern PURB    LINUXAPI LUsbAllocUrb(int iso_packets);
extern void    LINUXAPI LUsbFreeUrb(PURB urb);
extern STATUS  LINUXAPI LUsbSubmitUrb(void* dev, PURB urb);

extern STATUS  LINUXAPI LUsbWriteVendorData(void* dev, LPVOID src, WORD len, BYTE request, WORD value, WORD index, int wait);
extern STATUS  LINUXAPI LUsbReadVendorData(void* dev, LPVOID dst, WORD len, BYTE request, WORD value, WORD index);

extern STATUS  LINUXAPI LUsbSetInterface(void* dev, int iface, int asetting);
extern STATUS  LINUXAPI LUsbStartDevice(void* dev, int iface, int asetting);
extern STATUS  LINUXAPI LUsbStopDevice(void* dev);

extern STATUS  LINUXAPI LUsbFindPipe(void* dev, int iface, int asetting, unsigned char addr, PUSBD_PIPE_INFORMATION pipe);


// Device-File
#define IRP_NO_MINOR                   0xFFFFFFFF

#define IRP_FIO_MASK                   0x00ff
#define IRP_PNP_MASK                   0xff00

#define IRP_MJ_CREATE                  0x0001
#define IRP_MJ_CLOSE                   0x0002
#define IRP_MJ_FLUSH                   0x0003
#define IRP_MJ_READ                    0x0004
#define IRP_MJ_WRITE                   0x0005
#define IRP_MJ_DEVICE_CONTROL          0x0007
#define IRP_MJ_INTERNAL_DEVICE_CONTROL 0x0008
#define IRP_MJ_CLEANUP                 0x0009
#define IRP_MJ_POLL                    0x000A

#define IRP_MJ_PNP                     0x0100
#define IRP_MJ_POWER                   0x0200

#define IRP_MN_REMOVE_DEVICE           0x00000001
#define IRP_MN_STOP_DEVICE             0x00000002
#define IRP_MN_SET_POWER               0x00000010
#define IRP_MN_QUERY_POWER             0x00000020

#define IRP_ARG_POWER_DOWN             0x01
#define IRP_ARG_POWER_UP               0x02

typedef struct _Params {
	LIST_ENTRY  entry;
	LPVOID		pIrpHdl;
	LPVOID		pFileHdl;
	LPBYTE		ioBuffer;
	STATUS		status;
	DWORD 		ioInLen;
	DWORD			ioOutLen;
	DWORD			ioCnt;
	DWORD			evt_maj;
	DWORD			evt_min;
	DWORD			minor;
	DWORD       wakeFlag;
	void        (*Complete)(struct _Params* irp);
	void*       Context;
	LONG        ref;
	} __attribute__((packed)) Params, *PParams, *LPParams;

extern STATUS	LINUXAPI LFioRegCharDrv(UINT major, LPCSTR name);
extern void		LINUXAPI LFioUnRegCharDrv(UINT major, LPCSTR name);

extern void    LINUXAPI LRegisterIoctl32Conversion(DWORD code);
extern void    LINUXAPI LUnregisterIoctl32Conversion(DWORD code);

extern void    LINUXAPI LCompleteRequest(Params* irp);

// Interrupt
#define IRQ_SHARED    1
#define IRQ_EXCLUSIVE 0

typedef void (*LIB_TASK)(LPVOID) LINUXAPI;
typedef BOOL (*LIB_IRQHDL)(int, LPVOID, LPVOID) LINUXAPI;

extern BOOL    LINUXAPI LIrqRequest(UINT irq, BOOL shared, LPCSTR descr, LPVOID ISRObj);
extern void    LINUXAPI LIrqFree(UINT irq, LPVOID ISRObj);
extern BOOL    LINUXAPI LInInterrupt(void);

// Enable/Disable Interrupts, SMPLock
extern DWORD   LINUXAPI LSMPLock(void);
extern void    LINUXAPI LSMPUnlock(DWORD cpu);

// Systemzeit
extern void    LINUXAPI LGetCurrentTime(short tim[6]);
extern void    LINUXAPI LGetCurrentTime2(unsigned int tim[2]);

// Interlocked... Funktionen
extern LONG    LINUXAPI InterlockedIncrement(LPLONG val);
extern LONG    LINUXAPI InterlockedDecrement(LPLONG val);

// Speicher
extern LPVOID  LINUXAPI LAllocateSystemMemory(DWORD size, BOOL lock);
extern void    LINUXAPI LFreeSystemMemory(LPVOID buf);
extern BOOL    LINUXAPI LPageAllocate(LPVOID pDev, DWORD size, LPDWORD physAddr, LPVOID* linAddr);
extern void    LINUXAPI LPageFree(LPVOID pDev, DWORD size, DWORD physAddr, LPVOID linAddr);

// Debug
#define OS_EMERG   "<0>"
#define OS_ERROR	 "<3>"
#define OS_WARNING "<4>"
#define OS_NOTICE	 "<5>"
#define OS_INFO	 "<6>"
#define OS_DEBUG	 "<7>"

extern void    LINUXAPI LDbgPrint(LPCSTR hdr, LPCSTR txt);
extern BOOL    LINUXAPI LDbgInInterrupt(void);

// diverses
extern DWORD   LINUXAPI LGetSystemTime(void);
extern BOOL    LINUXAPI LCheckForRoot(void);

// scheduling
extern void    LINUXAPI LDrvSleep(DWORD msec);

// Persistent storage
extern void    LINUXAPI LSetPersist(LPCSTR key, LPVOID buf, int len);
extern int     LINUXAPI LGetPersist(LPCSTR key, LPVOID buf, int len);
extern int     LINUXAPI LDumpFile(LPCSTR fnam, LPVOID buf, int len);

///////////////////////////////////////////////////////////////////////////////
// Waiter API
#define NO_TIMEOUT        -1

extern void  LINUXAPI LWaitableDestroy(void* obj);
extern int   LINUXAPI LWaitableWait(void* obj, long timeout);
extern void  LINUXAPI LWaitableWake(void* obj);

///////////////////////////////////////////////////////////////////////////////
// Mutex API
typedef struct __MUTEX* LMUTEX;

extern LMUTEX       LINUXAPI LMutexInit(void);

#define LMutexDestroy(l)          LWaitableDestroy(l)
#define LMutexLock(l, t)          LWaitableWait(l, t)
#define LMutexUnlock(l)           LWaitableWake(l)

///////////////////////////////////////////////////////////////////////////////
// Event API
typedef struct __EVENT* LEVENT;

extern LEVENT       LINUXAPI LEventInit(void);
extern void         LINUXAPI LEventClear(LEVENT evt);

#define LEventDestroy(e)          LWaitableDestroy(e)
#define LEventWait(e, t)          LWaitableWait(e, t)
#define LEventSet(e)              LWaitableWake(e)

///////////////////////////////////////////////////////////////////////////////
// Thread API
typedef struct __THREAD* LTHREAD;
typedef void (*LTHREADFUNC)(LTHREAD thrd, void* ctx) LINUXAPI;

extern LTHREAD      LINUXAPI LThreadInit(LTHREADFUNC func, void* ctx);

#define LThreadDestroy(thrd)      LWaitableDestroy(thrd)
#define LThreadJoin(thrd, t)      LWaitableWait(thrd, t)

///////////////////////////////////////////////////////////////////////////////
// Spinlock API
typedef struct __SPINLOCK* LSPINLOCK;

extern LSPINLOCK    LINUXAPI LSpinlockInit(void);
extern void         LINUXAPI LSpinlockAcquire(LSPINLOCK lock);
extern void         LINUXAPI LSpinlockRelease(LSPINLOCK lock);

#define LSpinlockDestroy(l)       LWaitableDestroy(l)

///////////////////////////////////////////////////////////////////////////////
// DPC API
typedef struct __DPC* LDPC;
typedef void (*LDPCFUNC)(LDPC dpc, void* ctx) LINUXAPI;

extern LDPC         LINUXAPI LDpcInit(LDPCFUNC func, LPVOID ctx);
extern void         LINUXAPI LDpcQueue(LDPC dpc);

#define LDpcDestroy(d)            LWaitableDestroy(d)

///////////////////////////////////////////////////////////////////////////////
// WaitQueue API
extern void*        LINUXAPI LWaitQueueInit(void);
extern void         LINUXAPI LWaitQueueDestroy(void* obj);
extern void         LINUXAPI LWaitQueueWait(void* obj, Params* irp);
extern void         LINUXAPI LWaitQueueWake(void* obj);

///////////////////////////////////////////////////////////////////////////////
// Socket API
struct _SOCKET;

typedef void (*LSOCKETCALLBACK)(struct _SOCKET* sock, void* ctx) LINUXAPI;

typedef struct _SOCKET {
	// User members
	void*            context;
	LSOCKETCALLBACK  onDataReady;
	LSOCKETCALLBACK  onDisconnect;

	// Private members
	void*            sock;
	void*            data_ready;
	void*            state_change;

	} LSOCKET;

typedef struct {
	short          sin_family;
	unsigned short sin_port;
	unsigned int   sin_addr;
	unsigned char  __pad[8];
	} LSOCKADDR;

extern int          LINUXAPI LSocketCreate(LSOCKET* sock, int domain, int type, int protocol);
extern int          LINUXAPI LSocketBind(LSOCKET* sock, LSOCKADDR* addr);
extern int          LINUXAPI LSocketListen(LSOCKET* sock, int backlog);
extern int          LINUXAPI LSocketConnect(LSOCKET* sock, LSOCKADDR* addr);
extern int          LINUXAPI LSocketAccept(LSOCKET* sock, LSOCKET* parent);
extern int          LINUXAPI LSocketRecvFrom(LSOCKET* sock, void* buf, unsigned int len, LSOCKADDR* addr);
extern int          LINUXAPI LSocketRecv(LSOCKET* sock, void* buf, unsigned int len);
extern int          LINUXAPI LSocketSendTo(LSOCKET* sock, void* buf, unsigned int len, LSOCKADDR* addr);
extern int          LINUXAPI LSocketSend(LSOCKET* sock, void* buf, unsigned int len);
extern int          LINUXAPI LSocketGetSockName(LSOCKET* sock, LSOCKADDR* addr);
extern int          LINUXAPI LSocketGetPeerName(LSOCKET* sock, LSOCKADDR* addr);
extern int          LINUXAPI LSocketSetSockOpt(LSOCKET* sock, int level, int optname, char* optval, unsigned int optlen);
extern int          LINUXAPI LSocketGetSockOpt(LSOCKET* sock, int level, int optname, char* optval, int* optlen);
extern int          LINUXAPI LSocketShutdown(LSOCKET* sock, int flags);
extern void         LINUXAPI LSocketClose(LSOCKET* sock);

// NetDevice Helper
extern void         LINUXAPI LRegisterNetworkNotifier(void);
extern void         LINUXAPI LUnregisterNetworkNotifier(void);

// Treiber Init- und Exit
extern LPVOID  LINUXAPI DriverAddDevice(PHARDW_DEV dev);
extern void    LINUXAPI DriverUnload(void);
extern STATUS  LINUXAPI DriverEntry(LPCSTR params);

// Zeiger auf IRPObject::_IrpHandler()
extern STATUS (*IRPHandler)(PParams) LINUXAPI;

#if defined(__cplusplus)
}
#endif

#endif

#endif
